"""Convert embeddings to various formats."""
import logging

import numpy as np
from tqdm import tqdm

import embeddix.utils.files as futils

__all__ = ('convert_to_numpy', 'convert_to_txt')

logger = logging.getLogger(__name__)


def convert_to_txt(embeddings_filepath, vocab_filepath):
    """Convert numpy embeddings to text."""
    logger.info('Converting input numpy file to txt: {}'
                .format(embeddings_filepath))
    vocab = futils.load_vocab(vocab_filepath)
    model = np.load(embeddings_filepath)
    txt_model_filepath = '{}.txt'.format(embeddings_filepath.split('.npy')[0])
    logger.info('Saving output to {}'.format(txt_model_filepath))
    with open(txt_model_filepath, 'w', encoding='utf-8') as otp:
        for word, idx in vocab.items():
            vector = ' '.join([str(item) for item in model[idx].tolist()])
            print('{} {}'.format(word, vector), file=otp)


def _convert_to_numpy(words, vectors, output_filepath):
    logger.info('Saving numpy vectors to {}.npy'.format(output_filepath))
    np.save(output_filepath, vectors)
    vocab_filepath = '{}.vocab'.format(output_filepath)
    logger.info('Saving vocabulary to {}'.format(vocab_filepath))
    with open(vocab_filepath, 'w', encoding='utf-8') as vocab_stream:
        for idx, word in enumerate(words):
            print('{}\t{}'.format(idx, word), file=vocab_stream)


def _extract_words_and_vectors_from_txt(embeddings_filepath):
    words = []
    vectors = None
    with open(embeddings_filepath, 'r', encoding='utf-8') as input_str:
        for line in tqdm(input_str,
                         total=futils.count_lines(embeddings_filepath)):
            line = line.strip()
            if line:
                tokens = line.split(' ', 1)
                key = tokens[0].lower()
                words.append(key)
                value = np.fromstring(tokens[1], dtype='float32', sep=' ')
                if not np.any(vectors):
                    vectors = value
                else:
                    vectors = np.vstack((vectors, value))
    return words, vectors


def convert_to_numpy(embeddings_filepath):
    """Convert text embeddings to numpy."""
    logger.info('Converting input txt file to numpy: {}'
                .format(embeddings_filepath))
    output_filepath = '{}'.format(embeddings_filepath.split('.txt')[0])
    words, vectors = _extract_words_and_vectors_from_txt(embeddings_filepath)
    _convert_to_numpy(words, vectors, output_filepath)



"""Evaluation metrics utilities."""

import logging
import numpy as np
import scipy.stats as stats

logger = logging.getLogger(__name__)

__all__ = ('purity', 'spearman', 'similarity', 'rmse', 'pearson', 'hmean',
           'energy')


# pylint: disable=C0103,W0622
def purity(y_true, y_pred):
    """Calculate purity for given true and predicted cluster labels.

    Parameters
    ----------
    y_true: array, shape: (n_samples, 1)
      True cluster labels
    y_pred: array, shape: (n_samples, 1)
      Cluster assingment.

    Returns
    -------
    purity: float
      Calculated purity.

    """
    assert len(y_true) == len(y_pred)
    true_clusters = np.zeros(shape=(len(set(y_true)), len(y_true)))
    pred_clusters = np.zeros_like(true_clusters)
    for id, cl in enumerate(set(y_true)):
        true_clusters[id] = (y_true == cl).astype('int')
    for id, cl in enumerate(set(y_pred)):
        pred_clusters[id] = (y_pred == cl).astype('int')
    M = pred_clusters.dot(true_clusters.T)
    return 1. / len(y_true) * np.sum(np.max(M, axis=1))


def hmean(values):
    """Return harmonic mean of values."""
    return stats.hmean(values)


def rmse(x, y):
    """Return root mean squared error."""
    return np.sqrt(((x - y) ** 2).mean())


# pylint: disable=C0103
def spearman(x, y):
    """Compute Spearman correlation."""
    return stats.spearmanr(x, y)[0]


def pearson(x, y):
    """Compute Pearson correlation."""
    return stats.pearsonr(x, y)[0]


def similarity(left_vectors, right_vectors):
    """Compute cosine similarity between two matrices."""
    if left_vectors.shape != right_vectors.shape:
        raise Exception(
            'Cannot compute similarity from numpy arrays of different shape: '
            '{} != {}'.format(left_vectors.shape(), right_vectors.shape()))
    dotprod = np.einsum('ij,ij->i', left_vectors, right_vectors)
    norms = np.linalg.norm(left_vectors, axis=1) * np.linalg.norm(
        right_vectors, axis=1)
    return dotprod / norms


def energy(matrix):
    """Compute the energy of the model.

    Energy is defined as the square of the Frobenius norm, which is also
    the square sum of singular values.
    """
    if matrix.ndim == 1:
        logger.info('Computing energy from singular values...')
        return np.sum(matrix**2)
    if matrix.__class__.__name__ != 'csr_matrix':
        raise Exception('Cannot compute energy on non-CSR sparse matrix')
    logger.info('Computing energy from squared Frobenius norm of scipy '
                'csr_matrix...')
    return matrix.power(2).sum()



"""Reduce size of embeddings by aligning their vocabularies."""
import os
import logging

import numpy as np

from scipy import sparse
from tqdm import tqdm

import embeddix.utils.files as futils

__all__ = ('_reduce_sparse', '_reduce_dense')

logger = logging.getLogger(__name__)


# pylint: disable=C0103
def _reduce_sparse(model, vocab, shared_vocab):
    """Reduce a sparse CSR matrix from vocab to shared_vocab (rows + columns)."""
    rows = []
    columns = []
    data = []
    model = model.tocoo()
    idx2word = {idx: word for word, idx in vocab.items()}
    for i, j, v in tqdm(zip(model.row, model.col, model.data),
                        total=len(model.data)):
        if idx2word[i] in shared_vocab and idx2word[j] in shared_vocab:
            rows.append(shared_vocab[idx2word[i]])
            columns.append(shared_vocab[idx2word[j]])
            data.append(v)
    return sparse.csr_matrix((data, (rows, columns)),
                             shape=(len(shared_vocab),
                                    len(shared_vocab)),
                             dtype='f')


def _reduce_dense(model, vocab, shared_vocab):
    """Reduce a dense ndarray from vocab to shared_vocab (rows only)."""
    _model = np.empty(shape=(len(shared_vocab), model.shape[1]))
    idx2word = {idx: word for word, idx in shared_vocab.items()}
    for idx, word in idx2word.items():
        _model[idx] = model[vocab[word]]
    return _model


# TODO: refactor?
def align_vocabs_and_models(embeddings_dirpath):
    """Align all models under dirpath on the same vocabulary."""
    logger.info('Aligning vocabularies under {}'
                .format(embeddings_dirpath))
    shared_vocab = futils.load_shared_vocab(embeddings_dirpath)
    logger.info('Shared vocabulary size = {}'.format(len(shared_vocab)))
    model_names = [filename.split('.npy')[0] for filename in
                   os.listdir(embeddings_dirpath) if filename.endswith('.npy')]
    logger.info('Processing models = {}'.format(model_names))
    for model_name in model_names:
        model_filepath = os.path.join(embeddings_dirpath,
                                      '{}.npy'.format(model_name))
        model = np.load(model_filepath)
        vocab_filepath = os.path.join(embeddings_dirpath,
                                      '{}.vocab'.format(model_name))
        vocab = futils.load_vocab(vocab_filepath)
        reduced_model = _reduce_dense(model, vocab, shared_vocab)
        reduced_model_filepath = os.path.join(embeddings_dirpath,
                                              '{}-reduced'.format(model_name))
        np.save(reduced_model_filepath, reduced_model)
        reduced_vocab_filepath = os.path.join(
            embeddings_dirpath, '{}-reduced.vocab'.format(model_name))
        with open(reduced_vocab_filepath, 'w', encoding='utf-8') as output_str:
            for word, idx in shared_vocab.items():
                print('{}\t{}'.format(idx, word), file=output_str)



"""Files utils."""
import logging

import numpy as np

from scipy import sparse

__all__ = ('save_vocab', 'load_vocab', 'count_lines',
           'load_sparse', 'save_sparse', 'save_dense', 'get_shared_vocab')

logger = logging.getLogger(__name__)


def load_dense(matrix_filepath):
    """Load numpy dense matrix."""
    return np.load(matrix_filepath)


def load_sparse(matrix_filepath):
    """Load scipy sparse matrix."""
    return sparse.load_npz(matrix_filepath)


def save_sparse(matrix_filepath, matrix):
    """Save scipy sparse matrix to file."""
    logger.info('Saving scipy sparse matrix to {}'.format(matrix_filepath))
    sparse.save_npz(matrix_filepath, matrix)


def save_dense(matrix_filepath, matrix):
    """Save numpy dense matrix to file."""
    logger.info('Saving numpy dense matrix to {}'.format(matrix_filepath))
    np.save(matrix_filepath, matrix)


def save_vocab(vocab_filepath, vocab):
    """Save vocabulary to file."""
    logger.info('Saving vocabulary to {}'.format(vocab_filepath))
    with open(vocab_filepath, 'w', encoding='utf-8') as vocab_stream:
        for word, idx in vocab.items():
            print('{}\t{}'.format(idx, word), file=vocab_stream)


def load_vocab(vocab_filepath):
    """Load word_to_idx dict mapping from .vocab filepath."""
    word_to_idx = {}
    logger.info('Loading vocabulary from {}'.format(vocab_filepath))
    with open(vocab_filepath, 'r', encoding='utf-8') as input_stream:
        for line in input_stream:
            linesplit = line.strip().split('\t')
            word_to_idx[linesplit[1]] = int(linesplit[0])
    return word_to_idx


def count_lines(input_filepath):
    """Count number of non-empty lines in file."""
    counter = 0
    with open(input_filepath, 'r', encoding='utf-8') as input_str:
        for line in input_str:
            if line.strip():
                counter += 1
    return counter


def get_shared_vocab(vocab1, vocab2):
    """Return intersection of two vocabs."""
    shared_words = set(word for word in vocab1 if word in vocab2)
    return {word: idx for idx, word in enumerate(shared_words)}



"""Evaluate embeddings on various tasks."""
import logging

import numpy as np
from sklearn.cluster import KMeans

import embeddix.utils.files as futils
import embeddix.utils.metrix as metrix
import embeddix.utils.data as dutils

__all__ = ('evaluate_distributional_space', 'evaluate_word_similarity')

logger = logging.getLogger(__name__)


def _evaluate_concept_categorization(model, vocab, dataset):
    if dataset not in ['ap', 'battig', 'essli']:
        raise Exception('Invalid concept categorization dataset: {}'
                        .format(dataset))
    logger.info('Evaluating concept categorization on {}'.format(dataset))
    categories_to_words = dutils.load_dataset(dataset, vocab)
    categories = categories_to_words.keys()
    centroids = np.empty(shape=(len(categories), model.shape[1]))
    for idx, (_, words) in enumerate(categories_to_words.items()):
        word_idxx = [vocab[word] for word in words]
        centroids[idx] = np.mean(model[word_idxx, :], axis=0)
    category_words = [word for words in categories_to_words.values()
                      for word in words]
    category_words_idx = [vocab[word] for word in category_words]
    pred_clusters = KMeans(init=centroids, n_init=1,
                           n_clusters=len(categories)).fit_predict(
                               model[category_words_idx, :])
    true_clusters = np.array([idx for idx, words in
                              enumerate(categories_to_words.values())
                              for _ in words])
    purity = metrix.purity(true_clusters, pred_clusters)
    logger.info('Cluster purity = {}'.format(purity))
    return purity


def evaluate_word_similarity(model, vocab, dataset):
    """Evaluate lexical similarity/relatedness."""
    if dataset not in ['men', 'simlex', 'simverb']:
        raise Exception('Invalid similarity dataset: {}'.format(dataset))
    logger.info('Evaluating word similarity on {}'.format(dataset))
    left_idx, right_idx, sim = dutils.load_dataset(dataset, vocab)
    left_vectors = model[left_idx]
    right_vectors = model[right_idx]
    model_sim = metrix.similarity(left_vectors, right_vectors)
    spr = metrix.spearman(sim, model_sim)
    logger.info('Spearman correlation = {}'.format(spr))
    return spr


def evaluate_distributional_space(embeddings_filepath, vocab_filepath,
                                  dataset):
    """Evaluate embeddings on lexical similarity or concept categorization."""
    if dataset not in ['ap', 'battig', 'essli', 'men', 'simlex', 'simverb']:
        raise Exception('Unsupported dataset: {}'.format(dataset))
    model = np.load(embeddings_filepath)
    vocab = futils.load_vocab(vocab_filepath)
    if dataset in ['men', 'simlex', 'simverb']:
        return evaluate_word_similarity(model, vocab, dataset)
    return _evaluate_concept_categorization(model, vocab, dataset)



"""Welcome to embeddix.

This is the entry point of the application.
"""
import os

import argparse
import logging
import logging.config

import embeddix.core.converter as converter
import embeddix.core.extractor as extractor
import embeddix.core.evaluator as evaluator
import embeddix.core.reducer as reducer
import embeddix.utils.config as cutils


logging.config.dictConfig(
    cutils.load(
        os.path.join(os.path.dirname(__file__), 'logging', 'logging.yml')))

logger = logging.getLogger(__name__)


def _reduce(args):
    reducer.align_vocabs_and_models(args.embeddings)


def _convert(args):
    if args.to == 'numpy':
        if not args.embeddings.endswith('.txt'):
            raise Exception('Invalid input file: should be a text file '
                            'ending with .txt')
        converter.convert_to_numpy(args.embeddings)
    else:
        if not args.embeddings.endswith('.npy'):
            raise Exception('Invalid input file: should be a numpy file '
                            'ending with .npy')
        if not args.vocab:
            raise Exception('Converting to txt requires specifying the '
                            '--vocab parameter')
        converter.convert_to_txt(args.embeddings, args.vocab)


def _extract(args):
    extractor.extract_vocab(args.embeddings)


def _evaluate(args):
    logger.info('Loading distributional space from {}'.format(args.embeddings))
    evaluator.evaluate_distributional_space(args.embeddings, args.vocab,
                                            args.dataset)


def main():
    """Launch embeddix."""
    parser = argparse.ArgumentParser(prog='embeddix')
    subparsers = parser.add_subparsers()
    parser_extract = subparsers.add_parser(
        'extract', formatter_class=argparse.RawTextHelpFormatter,
        help='extract vocab from embeddings txt file')
    parser_extract.set_defaults(func=_extract)
    parser_extract.add_argument('-e', '--embeddings', required=True,
                                help='input embedding in txt format')
    parser_convert = subparsers.add_parser(
        'convert', formatter_class=argparse.RawTextHelpFormatter,
        help='convert embeddings to and from numpy and txt formats')
    parser_convert.set_defaults(func=_convert)
    parser_convert.add_argument('-t', '--to', choices=['numpy', 'txt'],
                                help='output format: numpy or text')
    parser_convert.add_argument('-v', '--vocab',
                                help='absolute path to vocabulary')
    parser_convert.add_argument('-e', '--embeddings', required=True,
                                help='absolute path to embeddings file')
    parser_evaluate = subparsers.add_parser(
        'evaluate', formatter_class=argparse.RawTextHelpFormatter,
        help='evaluate embeddings model on various intrinsic tasks')
    parser_evaluate.set_defaults(func=_evaluate)
    parser_evaluate.add_argument('-e', '--embeddings', required=True,
                                 help='absolute path to .npy embeddings')
    parser_evaluate.add_argument('-v', '--vocab', required=True,
                                 help='absolute path to .vocab file')
    parser_evaluate.add_argument('-d', '--dataset', required=True,
                                 choices=['ap', 'battig', 'essli', 'men',
                                          'simlex', 'simverb'],
                                 help='which dataset to evaluate on')
    args = parser.parse_args()
    args.func(args)



"""Statistical significance utils.

This is an adaptation of Philipp Stinger's implementation of
the Steiger's (1980) test available here:
https://github.com/psinger/CorrelationStats/blob/master/corrstats.py
"""
import numpy as np
from scipy import stats

__all__ = ('steiger_test_pval')


# pylint: disable=C0103
def steiger_test_pval(xy, xz, yz, n, twotailed=True):
    """Calculate statisticical significance between two dependent correlations.

    @param xy: correlation coefficient between x and y
    @param xz: correlation coefficient between x and z
    @param yz: correlation coefficient between y and z
    @param n: number of elements in x, y and z
    @param twotailed: whether to calculate a one or two tailed test
    @return: p-val
    """
    d = xy - xz
    determin = 1 - xy * xy - xz * xz - yz * yz + 2 * xy * xz * yz
    av = (xy + xz)/2
    cube = (1 - yz) * (1 - yz) * (1 - yz)
    t2 = d * np.sqrt((n - 1) * (1 + yz)/(((2 * (n - 1)/(n - 3)) * determin + av * av * cube)))
    p = 1 - stats.t.cdf(abs(t2), n - 3)
    if twotailed:
        p *= 2
    return p



"""Dataset utils."""
import os
import logging
import csv

from collections import defaultdict


__all__ = ('load_dataset')


logger = logging.getLogger(__name__)

DATASETS = {
    'ap': os.path.join(os.path.dirname(os.path.dirname(__file__)),
                       'resources', 'ap.csv'),
    'battig': os.path.join(os.path.dirname(os.path.dirname(__file__)),
                           'resources', 'battig.csv'),
    'essli': os.path.join(os.path.dirname(os.path.dirname(__file__)),
                          'resources', 'essli-2008.csv'),
    'men': os.path.join(os.path.dirname(os.path.dirname(__file__)),
                        'resources', 'MEN_dataset_natural_form_full'),
    'simlex': os.path.join(os.path.dirname(os.path.dirname(__file__)),
                           'resources', 'SimLex-999.txt'),
    'simverb': os.path.join(os.path.dirname(os.path.dirname(__file__)),
                            'resources', 'SimVerb-3500.txt'),
}


def _load_categories_to_words_dict(dataset, vocab):
    if dataset not in ['ap', 'battig', 'essli']:
        raise Exception('Invalid concept categorization dataset: {}'
                        .format(dataset))
    categories_to_words = defaultdict(list)
    total_num_rows = 0
    with open(DATASETS[dataset], newline='') as csv_stream:
        next(csv_stream)
        reader = csv.reader(csv_stream, quotechar='|')
        for row in reader:
            if row[2]:
                total_num_rows += 1
                if row[2] in vocab:
                    categories_to_words[row[1]].append(row[2])
                else:
                    logger.info('Word not in vocabulary: {}'.format(row[2]))
    logger.info('Processing {} words out of {} in {} categories'
                .format(sum(len(x) for x in categories_to_words.values()),
                        total_num_rows, len(categories_to_words)))
    return categories_to_words


# pylint: disable=C0103
def _load_idx_and_sim(left, right, sim, vocab, dataset):
    """Load discretized features and similarities for a given dataset.

    Will retain only words that are in the vocabulary.
    Will filter sim accordingly. Will normalize sim between [0, 1]
    """
    left_idx = []
    right_idx = []
    f_sim = []
    if dataset not in ['men', 'simlex', 'simverb']:
        raise Exception('Unsupported dataset {}'.format(dataset))
    for l, r, s in zip(left, right, sim):
        # if l not in vocab:
        #     raise Exception('Word {} not in {} dataset'
        #                     .format(l, dataset))
        # if r not in vocab:
        #     raise Exception('Word {} not in {} dataset'
        #                     .format(r, dataset))
        if l in vocab and r in vocab:
            left_idx.append(vocab[l])
            right_idx.append(vocab[r])
            if dataset == 'men':  # men has sim in [0, 50]
                f_sim.append(s/50)
            else:  # all other datasets have sim in [0, 10]
                f_sim.append(s/10)
    logger.info('Testing on {} pairs out of {}'
                .format(len(f_sim), len(sim)))
    ratio = (len(f_sim) / len(sim)) * 100
    logger.info('Test pairs ratio = {}%'.format(round(ratio, 2)))
    return left_idx, right_idx, f_sim


def _load_word_pairs_and_sim(dataset):
    """Load word pairs and similarity from a given dataset."""
    if dataset not in ['men', 'simlex', 'simverb']:
        raise Exception('Unsupported dataset {}'.format(dataset))
    left = []
    right = []
    sim = []
    with open(DATASETS[dataset], 'r', encoding='utf-8') as data_stream:
        for line in data_stream:
            line = line.rstrip('\n')
            items = line.split()
            left.append(items[0])
            right.append(items[1])
            if dataset == 'men':
                sim.append(float(items[2]))
            else:
                sim.append(float(items[3]))
    return left, right, sim


def load_dataset(dataset, vocab):
    """Load dataset for concept categorization or word similarity."""
    if dataset not in ['ap', 'battig', 'essli', 'men', 'simlex', 'simverb']:
        raise Exception('Unsupported dataset: {}'.format(dataset))
    if dataset in ['men', 'simlex', 'simverb', 'ws353']:
        left, right, sim = _load_word_pairs_and_sim(dataset)
        return _load_idx_and_sim(left, right, sim, vocab, dataset)
    return _load_categories_to_words_dict(dataset, vocab)



"""Exposed functions."""
from .utils.files import get_shared_vocab
from .utils.files import load_vocab
from .utils.files import count_lines
from .utils.files import load_sparse
from .utils.files import load_dense
from .utils.files import save_sparse
from .utils.files import save_dense
from .utils.files import save_vocab
from .utils.data import load_dataset
from .utils.metrix import similarity
from .utils.metrix import spearman
from .utils.metrix import pearson
from .utils.metrix import rmse
from .utils.metrix import hmean
from .utils.metrix import energy
from .utils.stats import steiger_test_pval
from .core.reducer import _reduce_dense as reduce_dense
from .core.reducer import _reduce_sparse as reduce_sparse
from .core.converter import convert_to_txt
from .core.evaluator import evaluate_word_similarity



"""Extract vocabulary from txt embeddings."""
import logging

from tqdm import tqdm

import embeddix.utils.files as futils

__all__ = ('extract_vocab')

logger = logging.getLogger(__name__)


def extract_vocab(embeddings_filepath):
    """Extract vocabulary from text embeddings."""
    words = []
    logger.info('Extracting vocabulary from {}'.format(embeddings_filepath))
    with open(embeddings_filepath, 'r', encoding='utf-8') as m_stream:
        for line in tqdm(m_stream,
                         total=futils.count_lines(embeddings_filepath)):
            line = line.strip()
            word = line.split(' ')[0]
            words.append(word)
    vocab_filepath = '{}.vocab'.format(embeddings_filepath.split('.txt')[0])
    with open(vocab_filepath, 'w', encoding='utf-8') as v_stream:
        for idx, word in enumerate(words):
            print('{}\t{}'.format(idx, word), file=v_stream)
    logger.info('Extracted {} words'.format(len(words)))
